import imp
from datetime import datetime
import psycopg2
import psycopg2.extras
from logzero import logger


class BgpRecorder():
    datetime_format = '%Y%m%d%H%M'
    table_name_prefix = "bgprib_"

    '''
    Common methods
    '''

    def __init__(self, db_host: str, db_port: int, db_name: str, db_user: str,  db_password: str) -> None:
        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name
        self.db_user = db_user
        self.db_password = db_password

    def get_db_connection(self) -> psycopg2.extensions.connection:
        con = psycopg2.connect(host=self.db_host, port=self.db_port,
                               dbname=self.db_name, user=self.db_user, password=self.db_password)
        return con

    '''
    Client methods
    '''

    def get_tables(self) -> list:
        sql = '''
        SELECT
        pg_class.relname
        FROM pg_stat_user_tables
        INNER JOIN pg_class ON pg_stat_user_tables.relname = pg_class.relname
        where pg_class.reltuples > 1
        order by relname asc;
        '''
        with self.get_db_connection() as con:
            with con.cursor() as cur:
                cur.execute(sql)
                rows = cur.fetchall()  # [('bgprib_202207080313',),...]
                return [row[0] for row in rows]

    def get_nearest_tablename(self, target_datetime: datetime) -> str:
        # get available tables
        table_names = self.get_tables()

        # convert table_name to datetime
        datetimes = [datetime.strptime(table_name.split("_")[1], self.datetime_format)
                     for table_name in table_names]

        # search nearenst
        nearrest_datetime = min(
            datetimes, key=lambda x: abs(x - target_datetime))

        table_name = f"{self.table_name_prefix}{nearrest_datetime.strftime(self.datetime_format)}"
        return table_name

    def get_routes_from_address(self, address: str, table_name: str) -> list:
        sql = f"select * from {table_name} where inet %s << {table_name}.nlri;"

        with self.get_db_connection() as con:
            with con.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
                cur.execute(sql, [address])
                rows = cur.fetchall()
                return [dict(row) for row in rows]

    def get_routes_from_address_and_datetime(self, address: str, target_datetime: datetime) -> list:
        table_name = self.get_nearest_tablename(
            target_datetime=target_datetime)
        return self.get_routes_from_address(address=address, table_name=table_name)

    def get_route_count(self, table_name: str) -> int:
        sql = f"SELECT count(*) from {table_name};"

        with self.get_db_connection() as con:
            with con.cursor() as cur:
                cur.execute(sql)
                count = cur.fetchone()[0]
                return count

    def is_table_exists(self, table_name: str) -> bool:
        try:
            # レコード0なら意味無し
            return self.get_route_count(table_name=table_name) > 0
        except Exception as e:
            return False

    def drop_table(self, table_name: str) -> bool:
        sql = f"DROP TABLE {table_name};"
        with self.get_db_connection() as con:
            with con.cursor() as cur:
                cur.execute(sql)
            con.commit()
        return True

    '''
    insert methods
    '''

    def __query_buildar(self, route_obj: dict, table_name: str) -> str:
        column_string = ",".join(route_obj.keys())
        holders = ["%s" for obj in route_obj.keys()]
        value_holder = ",".join(holders)

        sql = f'insert into {table_name}({column_string}) values({value_holder});'

        return sql

    def create_new_rib_table(self, table_name: str) -> bool:
        sql = f'''
        create table {table_name}
        (
            id SERIAL NOT NULL,
            time timestamp,
            path_id integer,
            type_name VARCHAR (64),
            sequence integer,
            from_ip inet,
            from_as integer,
            originated  timestamp,
            origin VARCHAR(32),
            aspath VARCHAR (256),
            nlri_type VARCHAR(32),
            nlri cidr,
            nexthop  inet,
            community VARCHAR (256),
            large_community VARCHAR (256)
        );
        '''
        try:
            with self.get_db_connection() as con:
                with con.cursor() as cur:
                    cur.execute(sql)
                con.commit()
        except psycopg2.errors.DuplicateTable as e:
            logger.warning(f"Table:{table_name} is already exists. ignore")
        except Exception as e:
            logger.error("Create Table Error")
            logger.error(e)
            return False
        return True

    def insert_route(self, route_obj: dict, table_name: str, con: psycopg2.extensions.connection = None):
        sql = self.__query_buildar(route_obj, table_name)
        try:
            if con is None:
                with self.get_db_connection() as con:
                    with con.cursor() as cur:
                        cur.execute(sql, list(route_obj.values()))
                    return True
            else:
                with con.cursor() as cur:
                    cur.execute(sql, list(route_obj.values()))
                return True
        except Exception as e:
            logger.error("DB Error. Can not insert record.")
            logger.error(e)
            return False



from logzero import logger
import pickledb
import argparse
from datetime import datetime
import os
import json
import pathlib
import time

from .bgprecorder import BgpRecorder
from . import util
from importlib import metadata

__version__ = metadata.version(__package__)

if __version__ == "0.0.0":
    __version__ = "UNKNOWN(localy build)"


def query():
    '''
    Client CLI tool
    '''
    parser = argparse.ArgumentParser(
        description='bgpquery: get BGP rib json from bgprecorder db')
    parser.add_argument('--version', version='%(prog)s ' + __version__,
                        action='version',
                        default=False)
    parser.add_argument('-a', '--address', type=str,
                        help='target address', required=True)
    parser.add_argument('-t', '--datetime', default=None, type=str,
                        help=f'target datetime. example: "200601021504"')
    parser.add_argument('-H', '--db_host', default=os.getenv("BGPRECORDER_DB_HOST", "localhost"), type=str,
                        help=f'db host. default: localhost or $BGPRECORDER_DB_HOST')
    parser.add_argument('-p', '--db_port', default=int(os.getenv("BGPRECORDER_DB_PORT", "5432")), type=int,
                        help=f'db port. default: 5432 or $BGPRECORDER_DB_PORT')
    parser.add_argument('-u', '--db_user', default=os.getenv("BGPRECORDER_DB_USER", "postgres"), type=str,
                        help=f'db user. default: postgres or $BGPRECORDER_DB_USER')
    parser.add_argument('-w', '--db_password', default=os.getenv("BGPRECORDER_DB_PASSWORD"), type=str,
                        help=f'db user. default: None or $BGPRECORDER_DB_PASSWORD ')
    parser.add_argument('-d', '--db_name', default=os.getenv("BGPRECORDER_DB_NAME", "bgprecorder"), type=str,
                        help=f'db user. default: bgprecorder or $BGPRECORDER_DB_RECORDER')

    args = parser.parse_args()

    address = args.address
    if not args.datetime:
        target_datetime = datetime.now()
    else:
        target_datetime = datetime.strptime(
            args.datetime, BgpRecorder.datetime_format)

    br = BgpRecorder(db_host=args.db_host, db_port=args.db_port,
                     db_name=args.db_name, db_user=args.db_user, db_password=args.db_password)

    routes = br.get_routes_from_address_and_datetime(
        address=address, target_datetime=target_datetime)

    matched_routes = util.longest_match(routes=routes)

    for route in matched_routes:
        print(json.dumps(route, default=util.json_serial_default))


def recorder():
    parser = argparse.ArgumentParser(
        description='bgprecord dump BGP MRT rib to DB')
    parser.add_argument('--version', version='%(prog)s ' + __version__,
                        action='version',
                        default=False)
    parser.add_argument('-H', '--db_host', default=os.getenv("BGPRECORDER_DB_HOST", "localhost"), type=str,
                        help=f'db host. default: localhost or $BGPRECORDER_DB_HOST')
    parser.add_argument('-p', '--db_port', default=int(os.getenv("BGPRECORDER_DB_PORT", "5432")), type=int,
                        help=f'db port. default: 5432 or $BGPRECORDER_DB_PORT')
    parser.add_argument('-u', '--db_user', default=os.getenv("BGPRECORDER_DB_USER", "postgres"), type=str,
                        help=f'db user. default: postgres or $BGPRECORDER_DB_USER')
    parser.add_argument('-w', '--db_password', default=os.getenv("BGPRECORDER_DB_PASSWORD"), type=str,
                        help=f'db password. default: None or $BGPRECORDER_DB_PASSWORD')
    parser.add_argument('-d', '--db_name', default=os.getenv("BGPRECORDER_DB_NAME", "bgprecorder"), type=str,
                        help=f'db name. default: bgprecorder or $BGPRECORDER_DB_RECORDER')
    parser.add_argument('-c', '--compress', default=bool(os.getenv("BGPRECORDER_COMPRESS", False)), type=bool,
                        help=f'compress MRT dump after import. default: False')
    parser.add_argument('-i', '--duration', default=int(os.getenv("BGPRECORDER_DURATION", default="3600")), type=int,
                        help=f'interval of recording (sec.) default: 3600 or $BGPRECORDER_DURATION')
    parser.add_argument('-f', '--mrt_dump_files', default=os.getenv(
        "BGPRECORDER_TARGET_FILES", default="./mrt/*.dump"), type=str,
        help=f'target MRT dumpfile match rule.  default: ./mrt/*.dump or $BGPRECORDER_DURATION')

    args = parser.parse_args()

    saved_file_cache = pickledb.load(os.getenv(
        "BGPRECORDER_CACHE_FILE", default="/tmp/bgprecorder.db"), True)  # auto dump
    args.mrt_dump_files = args.mrt_dump_files

    br = BgpRecorder(db_host=args.db_host, db_port=args.db_port,
                     db_name=args.db_name, db_user=args.db_user, db_password=args.db_password)

    while True:
        logger.info("Cycle started.")
        # get current dump files
        files = util.get_files(args.mrt_dump_files)

        for file in files:
            table_name = util.get_table_name_from_file_path(file_path=file)
            logger.info(f"Check file:{file}")

            # check valid flag
            if saved_file_cache.get(table_name):
                logger.info(f"Already recorded: {file}")
                if args.compress:
                    logger.info(f"file: {file} try to compress....")
                    if util.bzip2(file):
                        logger.info(f"file: {file} is compressed.")
                continue

            # set valid flag
            if br.is_table_exists(table_name=table_name):
                logger.info(f"Already registered: {file}")
                logger.info(f"set valid flag")
                saved_file_cache.set(table_name, True)
                continue

            # 正常に登録できてないor新規なので一旦flagをfalseにする．
            saved_file_cache.set(table_name, False)

            # exists check
            if not pathlib.Path(file).exists():
                logger.warning("File not found. ignore")
                continue
            # check emptyfile or not
            if not os.stat(file).st_size > 0:
                logger.warning("Empty file. ignore")
                continue

            logger.info(f"file: {file} found! try to record...")
            if __create_table_and_insert_route(br=br, file_path=file):
                logger.info(f"file: {file} is successfully recorded.")
                # validation check when next reconcile?
            else:
                logger.error(f"file: {file} could not be recorded.")

        logger.info(f"Cycle finished. sleep {args.duration} sec.")
        time.sleep(args.duration)


def __create_table_and_insert_route(br: BgpRecorder, file_path: str) -> bool:

    table_name = util.get_table_name_from_file_path(file_path=file_path)

    cmd = f"bgpdump -m {file_path}"

    # drop table if exists
    if br.is_table_exists(table_name):
        br.drop_table(table_name=table_name)

    # create table
    if not br.create_new_rib_table(table_name):
        logger.error("Can not create table. abort")
        return False

    # insert
    insert_successed = False
    with br.get_db_connection() as con:
        for line in util.localExecGetLines(cmd):
            route_obj = util.parse_bgpdump_record_to_route_obj(
                line.decode().strip())

            if not br.insert_route(route_obj=route_obj, table_name=table_name, con=con):
                logger.error("Insert Error")
                con.cancel()
                break
        else:
            con.commit()
            insert_successed = True
    return insert_successed


if __name__ == "__main__":
    recorder()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bgprecorder']

package_data = \
{'': ['*']}

install_requires = \
['logzero>=1.7.0,<2.0.0',
 'pickleDB>=0.9.2,<0.10.0',
 'psycopg2-binary>=2.9.3,<3.0.0']

entry_points = \
{'console_scripts': ['bgpquery = bgprecorder.cli:query',
                     'bgprecorder = bgprecorder.cli:recorder']}

setup_kwargs = {
    'name': 'bgprecorder',
    'version': '0.1.4',
    'description': 'BGP rib timeseries recording tool',
    'long_description': '# bgp route collector\n\nCreate a BGP RIB time series database from MRT format dump files.\n\n## how to deploy\n### docker\n1. Create docker-compose.yml with reference to docker-compose.sample.yml.\n2. Create .env\n```\nBGPRECORDER_DB_HOST=postgres\nBGPRECORDER_DB_PORT=5432\nBGPRECORDER_DB_NAME=bgprecorder\nBGPRECORDER_DB_USER=postgres\nBGPRECORDER_DB_PASSWORD=PASSWORD\n\n```\n3. run\n```\ndocker-compose up -d\n```\n\n\n### native install\nTBD\n\n## CLI tool\n\n### bgprecord\n\n```\n$ bgprecorder  -h \nusage: bgprecorder [-h] [-H DB_HOST] [-p DB_PORT] [-u DB_USER] [-w DB_PASSWORD] [-d DB_NAME] [-c COMPRESS] [-i DURATION] [-f MRT_DUMP_FILES]\n\nbgprecord dump BGP MRT rib to DB\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -H DB_HOST, --db_host DB_HOST\n                        db host. default: localhost or $BGPRECORDER_DB_HOST\n  -p DB_PORT, --db_port DB_PORT\n                        db port. default: 5432 or $BGPRECORDER_DB_PORT\n  -u DB_USER, --db_user DB_USER\n                        db user. default: postgres or $BGPRECORDER_DB_USER\n  -w DB_PASSWORD, --db_password DB_PASSWORD\n                        db password. default: None or $BGPRECORDER_DB_PASSWORD\n  -d DB_NAME, --db_name DB_NAME\n                        db name. default: bgprecorder or $BGPRECORDER_DB_RECORDER\n  -c COMPRESS, --compress COMPRESS\n                        compress MRT dump after import. default: False\n  -i DURATION, --duration DURATION\n                        interval of recording (sec.) default: 3600 or $BGPRECORDER_DURATION\n  -f MRT_DUMP_FILES, --mrt_dump_files MRT_DUMP_FILES\n                        target MRT dumpfile match rule. default: ./mrt/*.dump or $BGPRECORDER_DURATION\n```\n\n\n```\n$ bash misc/env.sh\n$ bgprecord \n```\n\n### bgpquery\n```\n$ bgpquery  -h \nusage: bgpquery [-h] -a ADDRESS [-t DATETIME] [-H DB_HOST] [-p DB_PORT] [-u DB_USER] [-w DB_PASSWORD] [-d DB_NAME]\n\nbgpquery: get BGP rib json from bgprecorder db\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -a ADDRESS, --address ADDRESS\n                        target address\n  -t DATETIME, --datetime DATETIME\n                        target datetime. example: "200601021504"\n  -H DB_HOST, --db_host DB_HOST\n                        db host. default: localhost or $BGPRECORDER_DB_HOST\n  -p DB_PORT, --db_port DB_PORT\n                        db port. default: 5432 or $BGPRECORDER_DB_PORT\n  -u DB_USER, --db_user DB_USER\n                        db user. default: postgres or $BGPRECORDER_DB_USER\n  -w DB_PASSWORD, --db_password DB_PASSWORD\n                        db user. default: None or $BGPRECORDER_DB_PASSWORD\n  -d DB_NAME, --db_name DB_NAME\n                        db user. default: bgprecorder or $BGPRECORDER_DB_RECORDER\n```\n\n```\n$ bash misc/env.sh\n$ bgpquery -a 3ffe::114  -d 202207131800  | jq\n{\n  "id": 13735,\n  "time": "2022-07-13T17:00:32",\n  "path_id": 3204,\n  "type_name": "TABLE_DUMP2_AP",\n  "sequence": null,\n  "from_ip": "2001:200:e00:300:dad::4",\n  "from_as": 0,\n  "originated": null,\n  "origin": "IGP",\n  "aspath": "400 300",\n  "nlri_type": null,\n  "nlri": "3ffe::/32",\n  "nexthop": "2001:db8::ace",\n  "community": "4690:64500",\n  "large_community": null\n}\n{\n  "id": 13736,\n  "time": "2022-07-13T17:00:32",\n  "path_id": 4634,\n  "type_name": "TABLE_DUMP2_AP",\n  "sequence": null,\n  "from_ip": "2001:200:e00:300:dad::5",\n  "from_as": 0,\n  "originated": null,\n  "origin": "IGP",\n  "aspath": "100 200 300",\n  "nlri_type": null,\n  "nlri": "3ffe::/32",\n  "nexthop": "2001:db8::beaf",\n  "community": "4690:64501",\n  "large_community": null\n}\n```\n\n',
    'author': 'yas-nyan',
    'author_email': 'yas-nyan@sfc.wide.ad.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/wide-vsix/bgprecorder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from datetime import datetime, date
import ipaddress
import subprocess

import glob
import pathlib

from .bgprecorder import BgpRecorder

'''



New Util


'''


def json_serial_default(obj):
    # 日付型の場合には、文字列に変換します
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    # 上記以外はサポート対象外.
    raise TypeError("Type %s not serializable" % type(obj))


def longest_match(routes: list) -> list:
    longest_prefix = 0
    for route in routes:
        prefixlen = ipaddress.ip_network(route["nlri"]).prefixlen
        if prefixlen >= longest_prefix:
            longest_prefix = prefixlen
    return [route for route in routes if ipaddress.ip_network(route["nlri"]).prefixlen == longest_prefix]


def localExec(cmd):
    proc = subprocess.run(
        cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    return proc.returncode == 0


def localExecCaptureOutput(cmd):
    proc = subprocess.run(
        cmd, shell=True, capture_output=True, text=True)
    return proc.stdout


def bzip2(filename, delete_src=True):
    delete_options = "" if delete_src else "-k"
    cmd = f"bzip2 {delete_options} {filename}"
    return localExec(cmd)


def localExecGetLines(cmd):
    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    while True:
        line = proc.stdout.readline()
        if line:
            yield line

        if not line and proc.poll() is not None:
            break


def parse_bgpdump_record_to_route_obj(bgpdump_record: str) -> dict:
    params = bgpdump_record.split("|")
    route_obj = {
        "time": datetime.fromtimestamp(int(params[1])),
        "path_id": params[6],
        "type_name": params[0],
        "aspath": params[7],
        # sequence
        "from_ip": params[3],
        "from_as": params[4],
        "origin": params[8],
        # originated
        # nlri_type
        "nlri": params[5],
        "nexthop": params[9],
        "community": params[12],
        # large_community
    }
    return route_obj


def get_files(match_rule: str) -> list:
    files = glob.glob(f"{match_rule}")
    return files


def get_table_name_from_file_path(file_path: str) -> str:
    prefix = BgpRecorder.table_name_prefix
    table_name_origin = pathlib.Path(
        file_path).stem.replace('bz2', '').replace('.', '').replace('dump', '')  # TODO refine
    return prefix + table_name_origin
